from argparse import ArgumentParser
import cv2
from data import Data, imread, imread_mono
import json
import numpy as np
import os


def create_camera_matrix(fov, size):
    """Create camera matrix for given field of view and image size.

    :param fov: Field of view in the larger dimension, in degrees.
    :param size: Image size (height, width).
    :return: 3-by-3 camera matrix K with given parameters.

    The camera matrix has following structure:
    [[f, 0, cx],
     [0, f, cy],
     [0, 0,  1]]
    """

    # TODO: Create camera matrix.
    camera_matrix = None

    return camera_matrix


def camera_field_of_view(camera_matrix, size):
    """Compute camera field of view from camera matrix K and image size.

    :param camera_matrix: 3-by-3 camera matrix K.
    :param size: Image size (height, width).
    :return: Horizontal and vertical field of viev.
    """
    # TODO: Compute field of view in both directions.
    fov_x = None
    fov_y = None

    return fov_x, fov_y


def opencv_size(shape):
    return tuple(shape[1::-1])


def calibrate(images, pattern, points):
    """Calibrate camera using set of images with calibration pattern.

    :param images: Sequences of images.
    :param pattern: Pattern size (width, height).
    :param points: 3D points coordinates in pattern frame.
    :return: Camera matrix K, distortion coefficients.
    """

    # TODO: Implement intrinsics calibration.
    camera_matrix = None
    distortion_coeffs = None

    return camera_matrix, distortion_coeffs


def remap(image, camera_matrix, distortion_coeffs, new_camera_matrix, new_size):
    """Remap camera image using given intrinsic parameters of source and
    target cameras using linear interpolation.

    :param image: Source camera image.
    :param camera_matrix: 3-by-3 source camera matrix.
    :param distortion_coeffs: Distortion coefficients.
    :param new_camera_matrix: 3-by-3 target camera matrix.
    :param new_size: New image size (height, width).
    :return: remapped camera image
    """

    # TODO: Render image for camera with given field of view and resolution.
    new_image = None

    return new_image


def save(path, camera_matrix, distortion_coeffs):
    intrinsics = {'camera_matrix': camera_matrix.tolist(),
                  'distortion_coeffs': distortion_coeffs.tolist()}
    with open(path, 'w', encoding='utf-8') as fp:
        json.dump(intrinsics, fp, indent=4, sort_keys=True)


def load(path):
    with open(path, 'r', encoding='utf-8') as f:
        intrinsics = json.load(f)
    return np.array(intrinsics['camera_matrix']), np.array(intrinsics['distortion_coeffs'])


def checkerboard_points(pattern, unit):
    if isinstance(unit, (list, tuple)):
        unit = np.array(unit)[None]
    x, y = np.meshgrid(*(np.arange(n).astype(np.float32) for n in pattern))
    pts = np.stack((x.flatten(), y.flatten(), np.zeros_like(x.flatten())))
    pts = (unit * pts.T).astype(np.float32)
    return pts


def main():
    parser = ArgumentParser()
    parser.add_argument('action', choices=['calibrate', 'remap'])
    parser.add_argument('calibration', type=str, help='Calibration file path (e.g., intrinsics.json).')
    parser.add_argument('--pattern', type=int, nargs=2, help='Checkerboard pattern size.')
    parser.add_argument('--unit', type=float, nargs='+', help='Checkerboard pattern unit size in meters.')
    parser.add_argument('--fov', type=float, help='Field of view in degrees for the larger dimension.')
    parser.add_argument('--alpha', type=float, help='Optimal camera matrix coverage.')
    parser.add_argument('--size', type=int, nargs=2, help='Camera image size (height, width).')
    parser.add_argument('images', type=str, nargs='+', help='One or more images to process.')
    args = parser.parse_args()

    if args.action == 'calibrate':
        assert args.pattern is not None and len(args.pattern) == 2
        assert args.unit is not None and len(args.unit) >= 1
        points = checkerboard_points(args.pattern, args.unit)
        camera_matrix, distortion_coeffs = calibrate(Data(args.images, imread_mono), args.pattern, points)
        save(args.calibration, camera_matrix, distortion_coeffs)
    elif args.action == 'remap':
        camera_matrix, distortion_coeffs = load(args.calibration)
        images = Data(args.images, imread)
        size = images[0].shape[:2]
        if args.size is not None:
            new_size = args.size
        else:
            new_size = size
        # Create new camera matrix, either from field of view or coverage ratio.
        if args.fov is not None:
            new_camera_matrix = create_camera_matrix(args.fov, new_size)
            fov = args.fov
        elif args.alpha is not None:
            new_camera_matrix \
                = cv2.getOptimalNewCameraMatrix(camera_matrix, distortion_coeffs, opencv_size(size),
                                                args.alpha, opencv_size(new_size))[0]
            fov = max(camera_field_of_view(new_camera_matrix, new_size))

        for i, image in enumerate(images):
            prefix, ext = os.path.splitext(images.src[i])
            remapped = remap(image, camera_matrix, distortion_coeffs, new_camera_matrix, new_size)
            path = '%s_%.0f_%ix%i%s' % (prefix, fov, *new_size, '.png')
            cv2.imwrite(path, remapped)


if __name__ == '__main__':
    main()
